from dataclasses import dataclass

from ssl import CertificateError, match_hostname
from typing import Optional, List, cast

import cryptography
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.x509 import ExtensionNotFound, ExtensionOID, Certificate, load_pem_x509_certificate, TLSFeature
from cryptography.x509.ocsp import load_der_ocsp_response, OCSPResponseStatus, OCSPResponse
from nassl._nassl import X509
from nassl.cert_chain_verifier import CertificateChainVerifier, CertificateChainVerificationFailed
import nassl.ocsp_response

from sslyze.plugins.certificate_info._certificate_utils import extract_dns_subject_alternative_names, get_common_names
from sslyze.plugins.certificate_info._symantec import SymantecDistructTester
from sslyze.plugins.certificate_info.trust_stores.trust_store import TrustStore


@dataclass(frozen=True)
class PathValidationResult:
    """The result of trying to validate a server's certificate chain using a specific trust store.

    Attributes:
        trust_stores: The trust store used for validation.
        verified_certificate_chain: The verified certificate chain returned by OpenSSL.
            Index 0 is the leaf certificate and the last element is the anchor/CA certificate from the trust store.
            Will be None if the validation failed or the verified chain could not be built.
            Each certificate is parsed using the cryptography module; documentation is available at
            https://cryptography.io/en/latest/x509/reference/#x-509-certificate-object.
        openssl_error_string: The result string returned by OpenSSL's validation function; None if validation was
            successful.
        was_validation_successful: Whether the certificate chain is trusted when using supplied the trust_stores.
    """

    trust_store: TrustStore
    verified_certificate_chain: Optional[List[Certificate]]
    openssl_error_string: Optional[str]

    @property
    def was_validation_successful(self) -> bool:
        return True if self.verified_certificate_chain else False


@dataclass(frozen=True)
class CertificateDeploymentAnalysisResult:
    """The result of analyzing a server's certificate to verify its validity.

    Any certificate available within the fields that follow is parsed as a ``Certificate`` object using the cryptography
    module; documentation is available at
    https://cryptography.io/en/latest/x509/reference.html?highlight=Certificate#cryptography.x509.Certificate

    Attributes:
        received_certificate_chain: The certificate chain sent by the server; index 0 is the leaf certificate.
        verified_certificate_chain: The verified certificate chain returned by OpenSSL for one of the trust stores
            packaged within SSLyze. Will be ``None`` if the validation failed with all of the available trust stores
            (Apple, Mozilla, etc.). This is essentially a shortcut to
            ``path_validation_result_list[0].verified_certificate_chain``.
        path_validation_results: The result of validating the server's
            certificate chain using each trust store that is packaged with SSLyze (Mozilla, Apple, etc.).
            If for a given trust store, the validation was successful, the verified certificate chain built by OpenSSL
            can be retrieved from the ``PathValidationResult``.
        leaf_certificate_subject_matches_hostname: ``True`` if the leaf certificate's Common Name or Subject Alternative
            Names match the server's hostname.
        leaf_certificate_is_ev: ``True`` if the leaf certificate is Extended Validation, according to Mozilla.
        leaf_certificate_has_must_staple_extension: ``True`` if the OCSP must-staple extension is present in the leaf
            certificate.
        leaf_certificate_signed_certificate_timestamps_count: The number of Signed Certificate
            Timestamps (SCTs) for Certificate Transparency embedded in the leaf certificate. ``None`` if the version of
            OpenSSL installed on the system is too old to be able to parse the SCT extension.
        received_chain_has_valid_order: ``True`` if the certificate chain returned by the server was sent in the right
            order. `None`` if any of the certificates in the chain could not be parsed.
        received_chain_contains_anchor_certificate: ``True`` if the server included the anchor/root
            certificate in the chain it sends back to clients. ``None`` if the verified chain could not be built.
        verified_chain_has_sha1_signature: ``True`` if any of the leaf or intermediate certificates are
            signed using the SHA-1 algorithm. ``None`` if the verified chain could not be built.
        verified_chain_has_legacy_symantec_anchor: ``True`` if the certificate chain contains a distrusted Symantec
            anchor
            (https://blog.qualys.com/ssllabs/2017/09/26/google-and-mozilla-deprecating-existing-symantec-certificates).
            ``None`` if the verified chain could not be built.
        ocsp_response: The OCSP response returned by the server. ``None`` if no response was sent by the server or if
            the scan was run through an HTTP proxy (the proxy will not forward the server's OCSP response). If present,
            the OCSP response is an ``OCSPResponse`` object parsed using the cryptography module; documentation is
            available at
            https://cryptography.io/en/latest/x509/ocsp.html?highlight=OCSPResponse#cryptography.x509.ocsp.OCSPResponse
        ocsp_response_is_trusted: ``True`` if the OCSP response is trusted using the Mozilla trust store.
            ``None`` if no OCSP response was sent by the server.

    """

    received_certificate_chain: List[Certificate]
    leaf_certificate_subject_matches_hostname: bool
    leaf_certificate_has_must_staple_extension: bool
    leaf_certificate_is_ev: bool
    leaf_certificate_signed_certificate_timestamps_count: Optional[int]
    received_chain_contains_anchor_certificate: Optional[bool]
    received_chain_has_valid_order: Optional[bool]

    path_validation_results: List[PathValidationResult]
    verified_chain_has_sha1_signature: Optional[bool]
    verified_chain_has_legacy_symantec_anchor: Optional[bool]

    ocsp_response: Optional[OCSPResponse]
    ocsp_response_is_trusted: Optional[bool]

    @property
    def verified_certificate_chain(self) -> Optional[List[Certificate]]:
        """Get one of the verified certificate chains if one was successfully built using any of the trust stores.
        """
        for path_result in self.path_validation_results:
            if path_result.was_validation_successful:
                return path_result.verified_certificate_chain
        return None

    @property
    def verified_certificate_chain_as_pem(self) -> Optional[List[str]]:
        if self.verified_certificate_chain is None:
            return None

        pem_certs = []
        for certificate in self.verified_certificate_chain:
            pem_certs.append(certificate.public_bytes(Encoding.PEM).decode("ascii"))
        return pem_certs

    @property
    def received_certificate_chain_as_pem(self) -> List[str]:
        pem_certs = []
        for certificate in self.received_certificate_chain:
            pem_certs.append(certificate.public_bytes(Encoding.PEM).decode("ascii"))
        return pem_certs


class CertificateDeploymentAnalyzer:
    """Utility class for analyzing a certificate chain as deployed on a specific server.

    Useful for checking a server's certificate chain without having to use the CertificateInfoPlugin.
    """

    def __init__(
        self,
        server_hostname: str,
        server_certificate_chain_as_pem: List[str],
        server_ocsp_response: Optional[nassl._nassl.OCSP_RESPONSE],
        trust_stores_for_validation: List[TrustStore],
    ) -> None:
        self.server_hostname = server_hostname
        self.server_certificate_chain_as_pem = server_certificate_chain_as_pem
        self.server_ocsp_response = server_ocsp_response
        self.trust_stores_for_validation = trust_stores_for_validation

    def perform(self) -> CertificateDeploymentAnalysisResult:
        received_certificate_chain = [
            load_pem_x509_certificate(pem_cert.encode("ascii"), backend=default_backend())
            for pem_cert in self.server_certificate_chain_as_pem
        ]
        leaf_cert = received_certificate_chain[0]

        # OCSP Must-Staple
        has_ocsp_must_staple = False
        try:
            tls_feature_ext = leaf_cert.extensions.get_extension_for_oid(ExtensionOID.TLS_FEATURE)
            tls_feature_value = cast(TLSFeature, tls_feature_ext.value)
            for feature_type in tls_feature_value:
                if feature_type == cryptography.x509.TLSFeatureType.status_request:
                    has_ocsp_must_staple = True
                    break
        except ExtensionNotFound:
            pass

        # Received chain order
        is_chain_order_valid: Optional[bool] = True
        previous_issuer = None
        for index, cert in enumerate(received_certificate_chain):
            try:
                current_subject = cert.subject
            except ValueError:
                # Cryptography could not parse the certificate https://github.com/nabla-c0d3/sslyze/issues/495
                is_chain_order_valid = None
                break

            if index > 0:
                # Compare the current subject with the previous issuer in the chain
                if current_subject != previous_issuer:
                    is_chain_order_valid = False
                    break
            try:
                previous_issuer = cert.issuer
            except KeyError:
                # Missing issuer; this is okay if this is the last cert
                previous_issuer = None
            except ValueError:
                # Cryptography could not parse the certificate https://github.com/nabla-c0d3/sslyze/issues/495
                is_chain_order_valid = None
                break

        # Check if the leaf certificate is Extended Validation
        is_leaf_certificate_ev = False
        for trust_store in self.trust_stores_for_validation:
            if trust_store.ev_oids is None:
                # We only have the EV OIDs for Mozilla - skip other stores
                continue

            is_leaf_certificate_ev = trust_store.is_certificate_extended_validation(leaf_cert)

        # Check for Signed Timestamps
        number_of_scts: Optional[int] = 0
        try:
            # Look for the x509 extension
            sct_ext = leaf_cert.extensions.get_extension_for_oid(ExtensionOID.PRECERT_SIGNED_CERTIFICATE_TIMESTAMPS)
            if isinstance(sct_ext.value, cryptography.x509.UnrecognizedExtension):
                # The version of OpenSSL on the system is too old and can't parse the SCT extension
                number_of_scts = None

            # Count the number of entries in the extension
            sct_ext_value = cast(cryptography.x509.PrecertificateSignedCertificateTimestamps, sct_ext.value)
            number_of_scts = len(sct_ext_value)
        except ExtensionNotFound:
            pass

        # Try to generate the verified certificate chain using each trust store
        all_path_validation_results = []
        for trust_store in self.trust_stores_for_validation:
            path_validation_result = _verify_certificate_chain(self.server_certificate_chain_as_pem, trust_store)
            all_path_validation_results.append(path_validation_result)

        # Keep one trust store that was able to build the verified chain to then run additional checks
        trust_store_that_can_build_verified_chain = None
        verified_certificate_chain = None

        # But first tort the path validation results so the same trust_store always get picked for a given server
        def sort_function(path_validation: PathValidationResult) -> str:
            return path_validation.trust_store.name.lower()

        all_path_validation_results.sort(key=sort_function)

        # Then keep a trust store with a verified chain
        for path_validation_result in all_path_validation_results:
            if path_validation_result.was_validation_successful:
                trust_store_that_can_build_verified_chain = path_validation_result.trust_store
                verified_certificate_chain = path_validation_result.verified_certificate_chain
                break

        # Check if the anchor was sent by the server
        has_anchor_in_certificate_chain = None
        if verified_certificate_chain:
            has_anchor_in_certificate_chain = verified_certificate_chain[-1] in received_certificate_chain

        # Check if a SHA1-signed certificate is in the chain
        # Root certificates can still be signed with SHA1 so we only check leaf and intermediate certificates
        has_sha1_in_certificate_chain = None
        if verified_certificate_chain:
            has_sha1_in_certificate_chain = False
            for cert in verified_certificate_chain[:-1]:
                if isinstance(cert.signature_hash_algorithm, hashes.SHA1):
                    has_sha1_in_certificate_chain = True
                    break

        # Check if this is a distrusted Symantec-issued chain
        verified_chain_has_legacy_symantec_anchor = None
        if verified_certificate_chain:
            symantec_distrust_timeline = SymantecDistructTester.get_distrust_timeline(verified_certificate_chain)
            verified_chain_has_legacy_symantec_anchor = True if symantec_distrust_timeline else False

        # Check the OCSP response if there is one
        is_ocsp_response_trusted = None
        final_ocsp_response = None
        if self.server_ocsp_response:
            # Parse the OCSP response returned by nassl
            final_ocsp_response = load_der_ocsp_response(self.server_ocsp_response.as_der_bytes())

            # Check if the OCSP response is trusted
            if (
                trust_store_that_can_build_verified_chain
                and final_ocsp_response.response_status == OCSPResponseStatus.SUCCESSFUL
            ):
                try:
                    nassl.ocsp_response.verify_ocsp_response(
                        self.server_ocsp_response, trust_store_that_can_build_verified_chain.path
                    )
                    is_ocsp_response_trusted = True
                except nassl.ocsp_response.OcspResponseNotTrustedError:
                    is_ocsp_response_trusted = False

        # All done
        return CertificateDeploymentAnalysisResult(
            received_certificate_chain=received_certificate_chain,
            leaf_certificate_subject_matches_hostname=_certificate_matches_hostname(leaf_cert, self.server_hostname),
            leaf_certificate_has_must_staple_extension=has_ocsp_must_staple,
            leaf_certificate_is_ev=is_leaf_certificate_ev,
            leaf_certificate_signed_certificate_timestamps_count=number_of_scts,
            received_chain_contains_anchor_certificate=has_anchor_in_certificate_chain,
            received_chain_has_valid_order=is_chain_order_valid,
            verified_chain_has_sha1_signature=has_sha1_in_certificate_chain,
            verified_chain_has_legacy_symantec_anchor=verified_chain_has_legacy_symantec_anchor,
            path_validation_results=all_path_validation_results,
            ocsp_response=final_ocsp_response,
            ocsp_response_is_trusted=is_ocsp_response_trusted,
        )


def _certificate_matches_hostname(certificate: Certificate, server_hostname: str) -> bool:
    """Verify that the certificate was issued for the given hostname.
    """
    # Extract the names from the certificate to create the properly-formatted dictionary
    try:
        cert_subject = certificate.subject
    except ValueError:
        # Cryptography could not parse the certificate https://github.com/nabla-c0d3/sslyze/issues/495
        return False

    certificate_names = {
        "subject": (tuple([("commonName", name) for name in get_common_names(cert_subject)]),),
        "subjectAltName": tuple([("DNS", name) for name in extract_dns_subject_alternative_names(certificate)]),
    }
    # CertificateError is raised on failure
    try:
        match_hostname(certificate_names, server_hostname)  # type: ignore
        return True
    except CertificateError:
        return False


def _verify_certificate_chain(server_certificate_chain: List[str], trust_store: TrustStore) -> PathValidationResult:
    server_chain_as_x509s = [X509(pem_cert) for pem_cert in server_certificate_chain]
    chain_verifier = CertificateChainVerifier.from_file(trust_store.path)

    verified_chain: Optional[List[Certificate]]
    try:
        openssl_verify_str = None
        verified_chain_as_509s = chain_verifier.verify(server_chain_as_x509s)
        verified_chain = [
            load_pem_x509_certificate(x509_cert.as_pem().encode("ascii"), backend=default_backend())
            for x509_cert in verified_chain_as_509s
        ]
    except CertificateChainVerificationFailed as e:
        verified_chain = None
        openssl_verify_str = e.openssl_error_string

    return PathValidationResult(
        trust_store=trust_store, verified_certificate_chain=verified_chain, openssl_error_string=openssl_verify_str
    )



import socket
import struct
from abc import abstractmethod, ABC
from enum import Enum
from typing import ClassVar, Optional


class ProtocolWithOpportunisticTlsEnum(str, Enum):
    """The list of plaintext protocols supported by SSLyze for opportunistic TLS upgrade (such as STARTTLS).

    This allows SSLyze to figure out how to complete an SSL/TLS handshake with the server.
    """

    SMTP = "SMTP"
    XMPP = "XMPP"
    XMPP_SERVER = "XMPP_SERVER"
    FTP = "FTP"
    POP3 = "POP3"
    LDAP = "LDAP"
    IMAP = "IMAP"
    RDP = "RDP"
    POSTGRES = "POSTGRES"

    @classmethod
    def from_default_port(cls, port: int) -> Optional["ProtocolWithOpportunisticTlsEnum"]:
        """Given a port number, return the protocol that uses this port number by default.
        """
        try:
            return _DEFAULT_PORTS[port]
        except KeyError:
            return None


_DEFAULT_PORTS = {
    587: ProtocolWithOpportunisticTlsEnum.SMTP,
    25: ProtocolWithOpportunisticTlsEnum.SMTP,
    5222: ProtocolWithOpportunisticTlsEnum.XMPP,
    5269: ProtocolWithOpportunisticTlsEnum.XMPP_SERVER,
    109: ProtocolWithOpportunisticTlsEnum.POP3,
    110: ProtocolWithOpportunisticTlsEnum.POP3,
    143: ProtocolWithOpportunisticTlsEnum.IMAP,
    220: ProtocolWithOpportunisticTlsEnum.IMAP,
    21: ProtocolWithOpportunisticTlsEnum.FTP,
    3268: ProtocolWithOpportunisticTlsEnum.LDAP,
    389: ProtocolWithOpportunisticTlsEnum.LDAP,
    3389: ProtocolWithOpportunisticTlsEnum.RDP,
    5432: ProtocolWithOpportunisticTlsEnum.POSTGRES,
}


class OpportunisticTlsError(Exception):
    pass


class _OpportunisticTlsHelper(ABC):
    @abstractmethod
    def prepare_socket_for_tls_handshake(self, sock: socket.socket) -> None:
        """Send the right protocol-specific requests to prepare the server for the TLS handshake.
        """
        pass


class _SmtpHelper(_OpportunisticTlsHelper):
    """Perform an SMTP StartTLS negotiation.
    """

    def prepare_socket_for_tls_handshake(self, sock: socket.socket) -> None:
        # Get the SMTP banner
        sock.recv(2048)

        # Send a EHLO and wait for the 250 status
        sock.send(b"EHLO sslyze.scan\r\n")
        if b"250 " not in sock.recv(2048):
            raise OpportunisticTlsError("SMTP EHLO was rejected")

        # Send a STARTTLS
        sock.send(b"STARTTLS\r\n")
        if b"220" not in sock.recv(2048):
            raise OpportunisticTlsError("SMTP STARTTLS not supported")


class _XmppHelper(_OpportunisticTlsHelper):
    """Perform an XMPP StartTLS negotiation.
    """

    XMPP_OPEN_STREAM = (
        "<stream:stream xmlns='jabber:client' xmlns:stream='http://etherx.jabber.org/streams' "
        "xmlns:tls='http://www.ietf.org/rfc/rfc2595.txt' to='{xmpp_to}' xml:lang='en' version='1.0'>"
    )
    XMPP_STARTTLS = b"<starttls xmlns='urn:ietf:params:xml:ns:xmpp-tls'/>"

    def __init__(self, xmpp_to: str) -> None:
        self._xmpp_to = xmpp_to

    def prepare_socket_for_tls_handshake(self, sock: socket.socket) -> None:
        # Open an XMPP stream
        sock.send(self.XMPP_OPEN_STREAM.format(xmpp_to=self._xmpp_to).encode("utf-8"))

        # Get the server's features and check for an error
        server_resp = sock.recv(4096)
        if b"<stream:error>" in server_resp:
            raise OpportunisticTlsError("Error opening XMPP stream, try --xmpp_to")
        elif b"</stream:features>" not in server_resp:
            # Get all the server features before initiating startTLS
            sock.recv(4096)

        # Send a STARTTLS message
        sock.send(self.XMPP_STARTTLS)
        xmpp_resp = sock.recv(2048)

        if b"host-unknown" in xmpp_resp:
            raise OpportunisticTlsError("Error opening XMPP stream: server returned host-unknown error, try --xmpp_to")

        if b"proceed" not in xmpp_resp:
            raise OpportunisticTlsError("XMPP STARTTLS not supported")


class _XmppServerHelper(_XmppHelper):
    XMPP_OPEN_STREAM = (
        "<stream:stream xmlns='jabber:server' xmlns:stream='http://etherx.jabber.org/streams' "
        "xmlns:tls='http://www.ietf.org/rfc/rfc2595.txt' to='{xmpp_to}' xml:lang='en' version='1.0'>"
    )


class _LdapHelper(_OpportunisticTlsHelper):
    """Performs an LDAP StartTLS negotiation.
    """

    START_TLS_CMD = b"0\x1d\x02\x01\x01w\x18\x80\x161.3.6.1.4.1.1466.20037"
    START_TLS_OK = b"\x30\x0c\x02\x01\x01\x78\x07\x0a\x01\x00\x04\x00\x04"
    START_TLS_OK2 = b"Start TLS request accepted"
    START_TLS_OK_APACHEDS = (
        b"\x30\x26\x02\x01\x01\x78\x21\x0a\x01\x00\x04\x00\x04\x00\x8a\x16\x31\x2e\x33\x2e\x36"
        b"\x2e\x31\x2e\x34\x2e\x31\x2e\x31\x34\x36\x36\x2e\x32\x30\x30\x33\x37\x8b\x00"
    )

    def prepare_socket_for_tls_handshake(self, sock: socket.socket) -> None:
        sock.send(self.START_TLS_CMD)
        data = sock.recv(2048)
        if self.START_TLS_OK not in data and self.START_TLS_OK_APACHEDS not in data and self.START_TLS_OK2 not in data:
            raise OpportunisticTlsError(f"LDAP AUTH TLS was rejected; returned: {repr(data)}")


class _RdpHelper(_OpportunisticTlsHelper):
    """Perform an RDP StartTLS negotiation.
    """

    ERR_NO_STARTTLS = "RDP AUTH TLS was rejected"

    START_TLS_CMD = b"\x03\x00\x00\x13\x0E\xE0\x00\x00\x00\x00\x00\x01\x00\x08\x00\x03\x00\x00\x00"
    START_TLS_OK = b"Start TLS request accepted."

    def prepare_socket_for_tls_handshake(self, sock: socket.socket) -> None:
        sock.send(self.START_TLS_CMD)
        data = sock.recv(4)
        if not data or len(data) != 4 or data[:2] != b"\x03\x00":
            raise OpportunisticTlsError(self.ERR_NO_STARTTLS)
        packet_len = struct.unpack(">H", data[2:])[0] - 4
        data = sock.recv(packet_len)

        if not data or len(data) != packet_len:
            raise OpportunisticTlsError(self.ERR_NO_STARTTLS)


class _GenericOpportunisticTlsHelper(_OpportunisticTlsHelper, ABC):
    # To be defined in subclasses
    ERR_NO_STARTTLS: ClassVar[str]  # Error to display to the user when StartTLS failed

    START_TLS_CMD: ClassVar[bytes]  # How to initiate StartTLS with the server
    START_TLS_OK: ClassVar[bytes]  # What is returned by the server when StartTLS was successful

    SHOULD_WAIT_FOR_SERVER_BANNER = True

    def prepare_socket_for_tls_handshake(self, sock: socket.socket) -> None:
        # Grab the banner
        if self.SHOULD_WAIT_FOR_SERVER_BANNER:
            sock.recv(2048)

        # Send Start TLS
        sock.send(self.START_TLS_CMD)
        if self.START_TLS_OK not in sock.recv(2048):
            raise OpportunisticTlsError(self.ERR_NO_STARTTLS)


class _ImapHelper(_GenericOpportunisticTlsHelper):

    ERR_NO_STARTTLS = "IMAP START TLS was rejected"

    START_TLS_CMD = b". STARTTLS\r\n"
    START_TLS_OK = b". OK"


class _Pop3Helper(_GenericOpportunisticTlsHelper):

    ERR_NO_STARTTLS = "POP START TLS was rejected"

    START_TLS_CMD = b"STLS\r\n"
    START_TLS_OK = b"+OK"


class _FtpHelper(_GenericOpportunisticTlsHelper):

    ERR_NO_STARTTLS = "FTP AUTH TLS was rejected"

    START_TLS_CMD = b"AUTH TLS\r\n"
    START_TLS_OK = b"234"


class _PostgresHelper(_GenericOpportunisticTlsHelper):

    ERR_NO_STARTTLS = "Postgres AUTH TLS was rejected"

    START_TLS_CMD = b"\x00\x00\x00\x08\x04\xD2\x16\x2F"
    START_TLS_OK = b"S"
    SHOULD_WAIT_FOR_SERVER_BANNER = False


_START_TLS_HELPER_CLASSES = {
    ProtocolWithOpportunisticTlsEnum.SMTP: _SmtpHelper,
    ProtocolWithOpportunisticTlsEnum.XMPP: _XmppHelper,
    ProtocolWithOpportunisticTlsEnum.XMPP_SERVER: _XmppServerHelper,
    ProtocolWithOpportunisticTlsEnum.POP3: _Pop3Helper,
    ProtocolWithOpportunisticTlsEnum.IMAP: _ImapHelper,
    ProtocolWithOpportunisticTlsEnum.FTP: _FtpHelper,
    ProtocolWithOpportunisticTlsEnum.LDAP: _LdapHelper,
    ProtocolWithOpportunisticTlsEnum.RDP: _RdpHelper,
    ProtocolWithOpportunisticTlsEnum.POSTGRES: _PostgresHelper,
}


def get_opportunistic_tls_helper(
    protocol: ProtocolWithOpportunisticTlsEnum, xmpp_to_hostname: Optional[str]
) -> _OpportunisticTlsHelper:
    helper_cls = _START_TLS_HELPER_CLASSES[protocol]
    if protocol not in [ProtocolWithOpportunisticTlsEnum.XMPP, ProtocolWithOpportunisticTlsEnum.XMPP_SERVER]:
        opportunistic_tls_helper = helper_cls()
    else:
        if xmpp_to_hostname is None:
            raise ValueError("Received None for xmpp_to_hostname")
        opportunistic_tls_helper = helper_cls(xmpp_to=xmpp_to_hostname)

    return opportunistic_tls_helper



from dataclasses import dataclass
from typing import List, Optional, Dict, Any, Tuple, Union

from sslyze.plugins.plugin_base import (
    ScanCommandResult,
    ScanCommandImplementation,
    ScanCommandExtraArgument,
    ScanJob,
    ScanCommandCliConnector,
    OptParseCliOption,
    ScanJobResult,
)
from sslyze.plugins.session_resumption._resumption_with_id import (
    resume_with_session_id,
    _ScanJobResultEnum,
    TlsResumptionSupportEnum,
    ServerOnlySupportsTls13,
)
from sslyze.plugins.session_resumption._resumption_with_ticket import resume_with_tls_ticket
from sslyze.server_connectivity import ServerConnectivityInfo


@dataclass(frozen=True)
class SessionResumptionSupportExtraArgument(ScanCommandExtraArgument):
    """Additional configuration for running the SESSION_RESUMPTION scan command.

    Attributes:
        number_of_resumptions_to_attempt: The number of session resumptions (both with Session IDs and TLS
            Tickets) that SSLyze should attempt. The default value is 5, but a higher value such as 100 can be used to
            get a more accurate measure of how often session resumption succeeds or fails with the server.
    """

    number_of_resumptions_to_attempt: int


@dataclass(frozen=True)
class SessionResumptionSupportScanResult(ScanCommandResult):
    """The result of testing a server for TLS 1.2 session resumption support, using Session IDs and TLS tickets.

    Attributes:
        session_id_resumption_result: The overall result of session ID resumption testing.
        session_id_attempted_resumptions_count: The total number of session ID resumptions that were attempted.
        session_id_successful_resumptions_count: The number of session ID resumptions that were successful.
        tls_ticket_resumption_result: The overall result of TLS ticket resumption testing.
        tls_ticket_attempted_resumptions_count: The total number of TLS ticket resumptions that were attempted.
        tls_ticket_successful_resumptions_count: The number of TLS ticket resumptions that were successful.

    """

    session_id_resumption_result: TlsResumptionSupportEnum
    session_id_attempted_resumptions_count: int
    session_id_successful_resumptions_count: int

    tls_ticket_resumption_result: TlsResumptionSupportEnum
    tls_ticket_attempted_resumptions_count: int
    tls_ticket_successful_resumptions_count: int


def _resumption_result_to_console_output(
    resumption_result: TlsResumptionSupportEnum, successful_resumptions_count: int, attempted_resumptions_count: int,
) -> str:
    if resumption_result == TlsResumptionSupportEnum.FULLY_SUPPORTED:
        resumption_support_txt = "OK - Supported"
    elif resumption_result == TlsResumptionSupportEnum.PARTIALLY_SUPPORTED:
        resumption_support_txt = "PARTIALLY SUPPORTED"
    elif resumption_result == TlsResumptionSupportEnum.NOT_SUPPORTED:
        resumption_support_txt = "NOT SUPPORTED"
    else:
        raise ValueError(f"Unexpected value: {resumption_result}")

    resum_rate_txt = (
        f"({successful_resumptions_count} successful resumptions" f" out of {attempted_resumptions_count} attempts)"
    )

    return f"{resumption_support_txt} {resum_rate_txt}"


class _SessionResumptionSupportCliConnector(
    ScanCommandCliConnector[SessionResumptionSupportScanResult, SessionResumptionSupportExtraArgument]
):

    _cli_option = "resum"
    _cli_description = "Test a server for TLS 1.2 session resumption support using session IDs and TLS tickets."

    @classmethod
    def get_cli_options(cls) -> List[OptParseCliOption]:
        scan_command_option = super().get_cli_options()
        scan_command_option.append(
            OptParseCliOption(
                option="resum_attempts",
                help="To be used with --resum. Number of session resumptions (both with Session IDs and TLS Tickets)"
                " that SSLyze should attempt."
                " The default value is 5, but a higher value such as 100 can be used to get a more accurate"
                " measure of how often session resumption succeeds or fails with the server.",
                action="store",
            )
        )
        return scan_command_option

    @classmethod
    def find_cli_options_in_command_line(
        cls, parsed_command_line: Dict[str, Union[None, bool, str]]
    ) -> Tuple[bool, Optional[SessionResumptionSupportExtraArgument]]:
        # Check if --resum was used
        is_scan_cmd_enabled, _ = super().find_cli_options_in_command_line(parsed_command_line)

        # Check if --resum_attempts was used
        extra_arguments = None
        try:
            resum_attempts = parsed_command_line["resum_attempts"]
            if resum_attempts:
                try:
                    resum_attempts_as_int = int(resum_attempts)
                    extra_arguments = SessionResumptionSupportExtraArgument(
                        number_of_resumptions_to_attempt=resum_attempts_as_int
                    )
                except ValueError:
                    raise ValueError(f'Supplied value for --resum_attempts is not an integer: "{resum_attempts}"')
        except KeyError:
            pass

        return is_scan_cmd_enabled, extra_arguments

    @classmethod
    def result_to_console_output(cls, result: SessionResumptionSupportScanResult) -> List[str]:
        result_as_txt = [cls._format_title("TLS 1.2 Session Resumption Support")]

        # Resumption with session IDs
        if result.session_id_resumption_result == TlsResumptionSupportEnum.SERVER_IS_TLS_1_3_ONLY:
            session_id_support_txt = "OK - Server only supports TLS 1.3 which doesn't support Session IDs"
        else:
            session_id_support_txt = _resumption_result_to_console_output(
                result.session_id_resumption_result,
                result.session_id_successful_resumptions_count,
                result.session_id_attempted_resumptions_count,
            )
        result_as_txt.append(f"      With Session IDs: {session_id_support_txt}.")

        # Resumption with TLS tickets
        if result.tls_ticket_resumption_result == TlsResumptionSupportEnum.SERVER_IS_TLS_1_3_ONLY:
            tls_ticket_support_txt = "OK - Server only supports TLS 1.3 which doesn't support TLS tickets"
        else:
            tls_ticket_support_txt = _resumption_result_to_console_output(
                result.tls_ticket_resumption_result,
                result.tls_ticket_successful_resumptions_count,
                result.tls_ticket_attempted_resumptions_count,
            )
        result_as_txt.append(f"      With TLS Tickets: {tls_ticket_support_txt}.")

        return result_as_txt


def _process_resumption_attempt_results(
    resumption_attempt_results: List[bool],
) -> Tuple[TlsResumptionSupportEnum, int, int]:
    total_attempts_count = len(resumption_attempt_results)
    successful_attempts_count = 0
    for was_resumption_successful in resumption_attempt_results:
        if was_resumption_successful:
            successful_attempts_count += 1

    if successful_attempts_count == 0:
        result = TlsResumptionSupportEnum.NOT_SUPPORTED
    elif successful_attempts_count == total_attempts_count:
        result = TlsResumptionSupportEnum.FULLY_SUPPORTED
    else:
        result = TlsResumptionSupportEnum.PARTIALLY_SUPPORTED

    return result, successful_attempts_count, total_attempts_count


class SessionResumptionSupportImplementation(ScanCommandImplementation[SessionResumptionSupportScanResult, None]):
    """Test a server for session resumption support using session IDs and TLS tickets.
    """

    cli_connector_cls = _SessionResumptionSupportCliConnector

    _DEFAULT_RESUMPTION_ATTEMPTS = 5

    @classmethod
    def scan_jobs_for_scan_command(
        cls,
        server_info: ServerConnectivityInfo,
        extra_arguments: Optional[SessionResumptionSupportExtraArgument] = None,
    ) -> List[ScanJob]:
        if extra_arguments:
            number_of_resumption_attempts = extra_arguments.number_of_resumptions_to_attempt
        else:
            number_of_resumption_attempts = cls._DEFAULT_RESUMPTION_ATTEMPTS

        # Test Session ID support
        session_id_scan_jobs = [
            ScanJob(function_to_call=resume_with_session_id, function_arguments=[server_info])
            for _ in range(number_of_resumption_attempts)
        ]

        # Test TLS tickets support
        tls_ticket_scan_jobs = [
            ScanJob(function_to_call=resume_with_tls_ticket, function_arguments=[server_info])
            for _ in range(number_of_resumption_attempts)
        ]

        return session_id_scan_jobs + tls_ticket_scan_jobs

    @classmethod
    def result_for_completed_scan_jobs(
        cls, server_info: ServerConnectivityInfo, scan_job_results: List[ScanJobResult]
    ) -> SessionResumptionSupportScanResult:
        if len(scan_job_results) == 0:
            raise RuntimeError(f"Unexpected number of scan jobs received: {scan_job_results}")

        # Sort TLS ticket VS session ID results
        results_dict: Dict[_ScanJobResultEnum, List[Any]] = {
            _ScanJobResultEnum.SESSION_ID_RESUMPTION: [],
            _ScanJobResultEnum.TLS_TICKET_RESUMPTION: [],
        }
        for job in scan_job_results:
            try:
                result_enum, value = job.get_result()
                results_dict[result_enum].append(value)
            except ServerOnlySupportsTls13:
                # If the server only supports TLS 1.3, none of the resumption mechanisms in this plugin are supported
                # by the server
                return SessionResumptionSupportScanResult(
                    session_id_resumption_result=TlsResumptionSupportEnum.SERVER_IS_TLS_1_3_ONLY,
                    session_id_attempted_resumptions_count=0,
                    session_id_successful_resumptions_count=0,
                    tls_ticket_resumption_result=TlsResumptionSupportEnum.SERVER_IS_TLS_1_3_ONLY,
                    tls_ticket_attempted_resumptions_count=0,
                    tls_ticket_successful_resumptions_count=0,
                )

        # Process session IDs resumption results
        session_id_result, session_id_successful_count, session_id_total_count = _process_resumption_attempt_results(
            results_dict[_ScanJobResultEnum.SESSION_ID_RESUMPTION]
        )

        # Process the TLS tickets resumption results
        tls_ticket_result, tls_ticket_successful_count, tls_ticket_total_count = _process_resumption_attempt_results(
            results_dict[_ScanJobResultEnum.TLS_TICKET_RESUMPTION]
        )

        # All done
        return SessionResumptionSupportScanResult(
            session_id_resumption_result=session_id_result,
            session_id_attempted_resumptions_count=session_id_total_count,
            session_id_successful_resumptions_count=session_id_successful_count,
            tls_ticket_resumption_result=tls_ticket_result,
            tls_ticket_attempted_resumptions_count=tls_ticket_total_count,
            tls_ticket_successful_resumptions_count=tls_ticket_successful_count,
        )



import queue
from traceback import TracebackException
from typing import List, Optional, Generator, Tuple, Sequence

from sslyze import ServerTlsProbingResult
from sslyze.errors import ConnectionToServerFailed
from sslyze.scanner._mass_connectivity_tester import MassConnectivityTester
from sslyze.scanner._mass_scanner import (
    MassScannerProducerThread,
    NoMoreServerScanRequestsSentinel,
)
from sslyze.scanner.models import (
    ServerScanRequest,
    ServerScanResult,
    ServerConnectivityStatusEnum,
    ServerScanStatusEnum,
)


from sslyze.scanner.scanner_observer import ScannerObserver


class Scanner:
    def __init__(
        self,
        per_server_concurrent_connections_limit: Optional[int] = None,
        concurrent_server_scans_limit: Optional[int] = None,
        observers: Optional[Sequence[ScannerObserver]] = None,
    ):
        self._observers: Sequence[ScannerObserver]
        if observers is None:
            self._observers = []
        else:
            self._observers = observers

        # Setup default values
        if per_server_concurrent_connections_limit is None:
            final_per_server_concurrent_connections_limit = 5
        else:
            final_per_server_concurrent_connections_limit = per_server_concurrent_connections_limit
        self._per_server_concurrent_connections_count = final_per_server_concurrent_connections_limit

        if concurrent_server_scans_limit is None:
            final_concurrent_server_scans_limit = 10
        else:
            final_concurrent_server_scans_limit = concurrent_server_scans_limit
        self._concurrent_server_scans_count = final_concurrent_server_scans_limit

        self._connectivity_tester = MassConnectivityTester(self._concurrent_server_scans_count)

    def queue_scans(self, server_scan_requests: List[ServerScanRequest]) -> None:
        if self._has_started_work:
            raise ValueError("Already submitted scan requests")

        if not server_scan_requests:
            raise ValueError("Submitted emtpy list of server_scan_requests")

        # Start with connectivity testing
        self._connectivity_tester.start_work(server_scan_requests)

    @property
    def _has_started_work(self) -> bool:
        return self._connectivity_tester.has_started_work

    def get_results(self) -> Generator[ServerScanResult, None, None]:
        if not self._has_started_work:
            raise ValueError("No scan requests have been submitted")

        # Setup the queues for running and completing the scans
        server_scan_requests_queue: "queue.Queue[Tuple[ServerScanRequest, ServerTlsProbingResult]]" = queue.Queue()
        server_scan_results_queue: "queue.Queue[ServerScanResult]" = queue.Queue()

        def server_connectivity_test_completed_callback(
            server_scan_request: ServerScanRequest, connectivity_result: ServerTlsProbingResult
        ) -> None:
            for inner_observer in self._observers:
                inner_observer.server_connectivity_test_completed(server_scan_request, connectivity_result)

            # Since the server is reachable, queue the actual scan commands
            server_scan_requests_queue.put((server_scan_request, connectivity_result))

        def server_connectivity_test_error_callback(
            server_scan_request: ServerScanRequest, connectivity_error: ConnectionToServerFailed
        ) -> None:
            for inner_observer in self._observers:
                inner_observer.server_connectivity_test_error(server_scan_request, connectivity_error)

            # Since the server is not reachable, there is nothing else to do
            server_scan_results_queue.put(
                ServerScanResult(
                    uuid=server_scan_request.uuid,
                    server_location=server_scan_request.server_location,
                    network_configuration=server_scan_request.network_configuration,
                    connectivity_status=ServerConnectivityStatusEnum.ERROR,
                    connectivity_error_trace=TracebackException.from_exception(connectivity_error),
                    connectivity_result=None,
                    scan_status=ServerScanStatusEnum.ERROR_NO_CONNECTIVITY,
                    scan_result=None,
                )
            )

        # Initialize the MassScanner for running scan commands
        mass_scanner_thread = MassScannerProducerThread(
            concurrent_server_scans_count=self._concurrent_server_scans_count,
            per_server_concurrent_connections_count=self._per_server_concurrent_connections_count,
            server_scan_requests_queue_in=server_scan_requests_queue,
            server_scan_results_queue_out=server_scan_results_queue,
        )
        mass_scanner_thread.start()

        # Wait until all connectivity testing has been completed
        self._connectivity_tester.wait_until_all_work_was_processed(
            server_connectivity_test_completed_callback=server_connectivity_test_completed_callback,
            server_connectivity_test_error_callback=server_connectivity_test_error_callback,
        )

        # Notify the MassScanner that all the scan requests have been queued
        server_scan_requests_queue.put(NoMoreServerScanRequestsSentinel())  # type: ignore

        # Wait for all scans to finish
        while True:
            server_scan_result = server_scan_results_queue.get(block=True)
            server_scan_results_queue.task_done()
            if isinstance(server_scan_result, NoMoreServerScanRequestsSentinel):
                # All scans have been completed
                break

            # Notify observers and yield the completed scan
            for observer in self._observers:
                observer.server_scan_completed(server_scan_result)

            yield server_scan_result

        # All done with the scans
        server_scan_requests_queue.join()
        server_scan_results_queue.join()
        mass_scanner_thread.join()

        for observer in self._observers:
            observer.all_server_scans_completed()



from abc import ABC
from typing import Optional

import pydantic

from sslyze.scanner.scan_command_attempt import ScanCommandAttemptStatusEnum, ScanCommandErrorReasonEnum


# Must be subclassed in order to add the result field
class ScanCommandAttemptAsJson(pydantic.BaseModel, ABC):
    class Config:
        orm_mode = True
        extra = "forbid"  # Fields must match between the JSON representation and the actual objects

    status: ScanCommandAttemptStatusEnum
    error_reason: Optional[ScanCommandErrorReasonEnum]
    error_trace: Optional[str]



from dataclasses import dataclass
from typing import Optional, Union

from nassl.ephemeral_key_info import EphemeralKeyInfo
from nassl.legacy_ssl_client import LegacySslClient
from nassl.ssl_client import ClientCertificateRequested, SslClient, BaseSslClient

from sslyze.errors import (
    ServerRejectedTlsHandshake,
    ServerTlsConfigurationNotSupported,
    TlsHandshakeTimedOut,
)
from sslyze.plugins.openssl_cipher_suites.cipher_suites import CipherSuite
from sslyze.server_connectivity import ServerConnectivityInfo, TlsVersionEnum
from sslyze.plugins.openssl_cipher_suites._tls12_workaround import WorkaroundForTls12ForCipherSuites


@dataclass(frozen=True)
class CipherSuiteAcceptedByServer:
    """
    ephemeral_key: The ephemeral key negotiated with the server when using (EC) DH cipher suites. None if the cipher
        suite does not use ephemeral keys or if the ephemeral key could not be retrieved.
    """

    cipher_suite: CipherSuite
    ephemeral_key: Optional[EphemeralKeyInfo]


@dataclass(frozen=True)
class CipherSuiteRejectedByServer:
    cipher_suite: CipherSuite
    error_message: str


def connect_with_cipher_suite(
    server_connectivity_info: ServerConnectivityInfo, tls_version: TlsVersionEnum, cipher_suite: CipherSuite
) -> Union[CipherSuiteAcceptedByServer, CipherSuiteRejectedByServer]:
    """Initiates a SSL handshake with the server using the SSL version and the cipher suite specified.
    """
    requires_legacy_openssl = True
    if tls_version == TlsVersionEnum.TLS_1_2:
        # For TLS 1.2, we need to pick the right version of OpenSSL depending on which cipher suite
        requires_legacy_openssl = WorkaroundForTls12ForCipherSuites.requires_legacy_openssl(cipher_suite.openssl_name)
    elif tls_version == TlsVersionEnum.TLS_1_3:
        requires_legacy_openssl = False

    ssl_connection = server_connectivity_info.get_preconfigured_tls_connection(
        override_tls_version=tls_version, should_use_legacy_openssl=requires_legacy_openssl
    )
    _set_cipher_suite_string(tls_version, cipher_suite.openssl_name, ssl_connection.ssl_client)

    ephemeral_key = None
    try:
        # Perform the SSL handshake
        ssl_connection.connect()
        ephemeral_key = ssl_connection.ssl_client.get_ephemeral_key()

    except ServerTlsConfigurationNotSupported:
        # SSLyze rejected the handshake because the server's DH config was too insecure; this means the
        # cipher suite is actually supported
        pass

    except ClientCertificateRequested:
        # When the handshake failed due to ClientCertificateRequested
        ephemeral_key = ssl_connection.ssl_client.get_ephemeral_key()
        pass

    except ServerRejectedTlsHandshake as e:
        return CipherSuiteRejectedByServer(cipher_suite=cipher_suite, error_message=e.error_message)

    except TlsHandshakeTimedOut as e:
        # Sometimes triggered by servers that don't support (at all) a specific version of TLS
        # Amazon Cloudfront does that with TLS 1.3
        # There's no easy way to differentiate this error from a network glitch/timeout
        return CipherSuiteRejectedByServer(cipher_suite=cipher_suite, error_message=e.error_message)

    finally:
        ssl_connection.close()

    return CipherSuiteAcceptedByServer(cipher_suite=cipher_suite, ephemeral_key=ephemeral_key)


def _set_cipher_suite_string(tls_version: TlsVersionEnum, cipher_suite_str: str, ssl_client: BaseSslClient) -> None:
    # Only enable the cipher suite to test; not trivial anymore since OpenSSL 1.1.1 and TLS 1.3
    if isinstance(ssl_client, SslClient):
        # With the modern OpenSSL client we have to manage TLS 1.3-specific cipher functions
        if tls_version == TlsVersionEnum.TLS_1_3:
            ssl_client.set_ciphersuites(cipher_suite_str)  # TLS 1.3 method
        else:
            ssl_client.set_cipher_list(cipher_suite_str)  # Legacy method
    elif isinstance(ssl_client, LegacySslClient):
        # With the legacy OpenSSL client, nothing special to do
        ssl_client.set_cipher_list(cipher_suite_str)
    else:
        raise RuntimeError("Should never happen")



import json
from enum import Enum
from pathlib import Path
from typing import Optional, Set, Dict

import pydantic
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from nassl.ephemeral_key_info import EcDhEphemeralKeyInfo, DhEphemeralKeyInfo

from sslyze import (
    ServerScanResult,
    ServerScanStatusEnum,
    ScanCommand,
    ScanCommandAttemptStatusEnum,
    CertificateInfoScanResult,
    AllScanCommandsAttempts,
    CipherSuitesScanResult,
    RobotScanResultEnum,
    SupportedEllipticCurvesScanResult,
)


class _MozillaCiphersAsJson(pydantic.BaseModel):
    caddy: Set[str]
    go: Set[str]
    iana: Set[str]
    openssl: Set[str]


class _MozillaTlsConfigurationAsJson(pydantic.BaseModel):
    certificate_curves: Set[str]
    certificate_signatures: Set[str]
    certificate_types: Set[str]
    ciphersuites: Set[str]
    ciphers: _MozillaCiphersAsJson
    dh_param_size: Optional[int]
    ecdh_param_size: int
    hsts_min_age: int
    maximum_certificate_lifespan: int
    ocsp_staple: bool
    recommended_certificate_lifespan: int
    rsa_key_size: Optional[int]
    server_preferred_order: bool
    tls_curves: Set[str]
    tls_versions: Set[str]


class _AllMozillaTlsConfigurationsAsJson(pydantic.BaseModel):
    modern: _MozillaTlsConfigurationAsJson
    intermediate: _MozillaTlsConfigurationAsJson
    old: _MozillaTlsConfigurationAsJson


class _MozillaTlsProfileAsJson(pydantic.BaseModel):
    version: str
    href: str
    configurations: _AllMozillaTlsConfigurationsAsJson


class MozillaTlsConfigurationEnum(str, Enum):
    MODERN = "modern"
    INTERMEDIATE = "intermediate"
    OLD = "old"


class ServerNotCompliantWithMozillaTlsConfiguration(Exception):
    def __init__(
        self, mozilla_config: MozillaTlsConfigurationEnum, issues: Dict[str, str],
    ):
        self.mozilla_config = mozilla_config
        self.issues = issues


class ServerScanResultIncomplete(Exception):
    """The server scan result does not have enough information to check it against Mozilla's configuration.
    """


SCAN_COMMANDS_NEEDED_BY_MOZILLA_CHECKER: Set[ScanCommand] = {
    ScanCommand.SSL_2_0_CIPHER_SUITES,
    ScanCommand.SSL_3_0_CIPHER_SUITES,
    ScanCommand.TLS_1_0_CIPHER_SUITES,
    ScanCommand.TLS_1_1_CIPHER_SUITES,
    ScanCommand.TLS_1_2_CIPHER_SUITES,
    ScanCommand.TLS_1_3_CIPHER_SUITES,
    ScanCommand.HEARTBLEED,
    ScanCommand.ROBOT,
    ScanCommand.OPENSSL_CCS_INJECTION,
    ScanCommand.TLS_COMPRESSION,
    ScanCommand.SESSION_RENEGOTIATION,
    ScanCommand.CERTIFICATE_INFO,
    ScanCommand.ELLIPTIC_CURVES,
}


class MozillaTlsConfigurationChecker:
    def __init__(self, mozilla_tls_profile: _MozillaTlsProfileAsJson):
        self._mozilla_tls_profile = mozilla_tls_profile

    @classmethod
    def get_default(cls) -> "MozillaTlsConfigurationChecker":
        json_profile_path = Path(__file__).parent.absolute() / "5.6.json"
        json_profile_as_str = json_profile_path.read_text()
        parsed_profile = _MozillaTlsProfileAsJson(**json.loads(json_profile_as_str))
        return cls(parsed_profile)

    def check_server(self, against_config: MozillaTlsConfigurationEnum, server_scan_result: ServerScanResult,) -> None:
        # Ensure the scan was successful
        if server_scan_result.scan_status != ServerScanStatusEnum.COMPLETED:
            raise ServerScanResultIncomplete("The server scan was not completed.")

        # Ensure all the scan command we need were run successfully
        for scan_command in SCAN_COMMANDS_NEEDED_BY_MOZILLA_CHECKER:
            scan_cmd_attempt = getattr(server_scan_result.scan_result, scan_command.value)
            if scan_cmd_attempt.status != ScanCommandAttemptStatusEnum.COMPLETED:
                raise ServerScanResultIncomplete(f"The {scan_command.value} result is missing.")

        # Now let's check the server's scan results against the Mozilla config
        mozilla_config: _MozillaTlsConfigurationAsJson = getattr(
            self._mozilla_tls_profile.configurations, against_config.value
        )
        all_issues: Dict[str, str] = {}

        # Checks on the certificate
        assert server_scan_result.scan_result
        assert server_scan_result.scan_result.certificate_info
        assert server_scan_result.scan_result.certificate_info.result
        issues_with_certificates = _check_certificates(
            cert_info_result=server_scan_result.scan_result.certificate_info.result, mozilla_config=mozilla_config,
        )
        all_issues.update(issues_with_certificates)

        # Checks on the TLS versions and cipher suites
        assert server_scan_result.scan_result
        issues_with_tls_ciphers = _check_tls_versions_and_ciphers(server_scan_result.scan_result, mozilla_config)
        all_issues.update(issues_with_tls_ciphers)

        # Checks on the TLS curves
        assert server_scan_result.scan_result.elliptic_curves.result
        issues_with_tls_curves = _check_tls_curves(
            server_scan_result.scan_result.elliptic_curves.result, mozilla_config,
        )
        all_issues.update(issues_with_tls_curves)

        # Checks on TLS vulnerabilities
        issues_with_tls_vulns = _check_tls_vulnerabilities(server_scan_result.scan_result)
        all_issues.update(issues_with_tls_vulns)

        if all_issues:
            raise ServerNotCompliantWithMozillaTlsConfiguration(
                mozilla_config=against_config, issues=all_issues,
            )


def _check_tls_curves(
    tls_curves_result: SupportedEllipticCurvesScanResult, mozilla_config: _MozillaTlsConfigurationAsJson,
) -> Dict[str, str]:
    issues_with_tls_curves = {}
    if tls_curves_result.supported_curves:
        supported_curves = {curve.name for curve in tls_curves_result.supported_curves}
    else:
        supported_curves = set()

    tls_curves_difference = supported_curves - mozilla_config.tls_curves
    if tls_curves_difference:
        issues_with_tls_curves[
            "tls_curves"
        ] = f"TLS curves {tls_curves_difference} are supported, but should be rejected."

    # TODO(AD): Disable the check on the curves; not even Google, Mozilla nor Cloudflare are compliant...
    # return problems_with_tls_curves
    return {}


def _check_tls_vulnerabilities(scan_result: AllScanCommandsAttempts) -> Dict[str, str]:
    issues_with_tls_vulns = {}
    assert scan_result.tls_compression.result
    if scan_result.tls_compression.result.supports_compression:
        issues_with_tls_vulns["tls_vulnerability_compression"] = "Server is vulnerable to TLS compression attacks."

    assert scan_result.openssl_ccs_injection.result
    if scan_result.openssl_ccs_injection.result.is_vulnerable_to_ccs_injection:
        issues_with_tls_vulns[
            "tls_vulnerability_ccs_injection"
        ] = "Server is vulnerable to the OpenSSL CCS injection attack."

    assert scan_result.heartbleed.result
    if scan_result.heartbleed.result.is_vulnerable_to_heartbleed:
        issues_with_tls_vulns["tls_vulnerability_heartbleed"] = "Server is vulnerable to the OpenSSL Heartbleed attack."

    assert scan_result.robot.result
    if scan_result.robot.result.robot_result == RobotScanResultEnum.VULNERABLE_STRONG_ORACLE:
        issues_with_tls_vulns["tls_vulnerability_robot"] = "Server is vulnerable to the ROBOT attack."

    assert scan_result.session_renegotiation.result
    if not scan_result.session_renegotiation.result.supports_secure_renegotiation:
        issues_with_tls_vulns[
            "tls_vulnerability_renegotiation"
        ] = "Server is vulnerable to the insecure renegotiation attack."

    return issues_with_tls_vulns


def _check_tls_versions_and_ciphers(
    scan_result: AllScanCommandsAttempts, mozilla_config: _MozillaTlsConfigurationAsJson,
) -> Dict[str, str]:
    # First parse the results related to TLS versions and ciphers
    tls_versions_supported = set()
    cipher_suites_supported = set()
    tls_1_3_cipher_suites_supported = set()
    curves_supported = set()
    smallest_ecdh_param_size = 100000
    smallest_dh_param_size = 100000
    for field_name, tls_version_name in [
        ("ssl_2_0_cipher_suites", "SSLv2"),
        ("ssl_3_0_cipher_suites", "SSLv3"),
        ("tls_1_0_cipher_suites", "TLSv1"),
        ("tls_1_1_cipher_suites", "TLSv1.1"),
        ("tls_1_2_cipher_suites", "TLSv1.2"),
        ("tls_1_3_cipher_suites", "TLSv1.3"),
    ]:
        tls_scan_result: CipherSuitesScanResult = getattr(scan_result, field_name).result
        if tls_scan_result.is_tls_version_supported:
            tls_versions_supported.add(tls_version_name)
            for accepted_cipher_suite in tls_scan_result.accepted_cipher_suites:
                if tls_version_name == "TLSv1.3":
                    tls_1_3_cipher_suites_supported.add(accepted_cipher_suite.cipher_suite.name)
                else:
                    cipher_suites_supported.add(accepted_cipher_suite.cipher_suite.name)

                ephemeral_key = accepted_cipher_suite.ephemeral_key
                if isinstance(ephemeral_key, EcDhEphemeralKeyInfo):
                    curves_supported.add(ephemeral_key.curve_name)
                    actual_key_size = ephemeral_key.size + 3  # OpenSSL returns 253 instead of 255 for the secret key
                    smallest_ecdh_param_size = min([smallest_ecdh_param_size, actual_key_size])

                elif isinstance(ephemeral_key, DhEphemeralKeyInfo):
                    smallest_dh_param_size = min([smallest_dh_param_size, ephemeral_key.size])

    # Then check the results
    issues_with_tls_ciphers = {}
    tls_versions_difference = tls_versions_supported - mozilla_config.tls_versions
    if tls_versions_difference:
        issues_with_tls_ciphers[
            "tls_versions"
        ] = f"TLS versions {tls_versions_difference} are supported, but should be rejected."

    tls_1_3_cipher_suites_difference = tls_1_3_cipher_suites_supported - mozilla_config.ciphersuites
    if tls_1_3_cipher_suites_difference:
        issues_with_tls_ciphers[
            "ciphersuites"
        ] = f"TLS 1.3 cipher suites {tls_1_3_cipher_suites_difference} are supported, but should be rejected."

    cipher_suites_difference = cipher_suites_supported - mozilla_config.ciphers.iana
    if cipher_suites_difference:
        issues_with_tls_ciphers[
            "ciphers"
        ] = f"Cipher suites {cipher_suites_difference} are supported, but should be rejected."

    if mozilla_config.ecdh_param_size and smallest_ecdh_param_size < mozilla_config.ecdh_param_size:
        issues_with_tls_ciphers["ecdh_param_size"] = (
            f"ECDH parameter size is {smallest_ecdh_param_size},"
            f" should be superior or equal to {mozilla_config.ecdh_param_size}."
        )

    if mozilla_config.dh_param_size and smallest_dh_param_size < mozilla_config.dh_param_size:
        issues_with_tls_ciphers["dh_param_size"] = (
            f"DH parameter size is {smallest_dh_param_size},"
            f" should be superior or equal to {mozilla_config.dh_param_size}."
        )

    return issues_with_tls_ciphers


def _check_certificates(
    cert_info_result: CertificateInfoScanResult, mozilla_config: _MozillaTlsConfigurationAsJson,
) -> Dict[str, str]:
    issues_with_certificates = {}
    deployed_key_algorithms = set()
    deployed_signature_algorithms = set()
    for cert_deployment in cert_info_result.certificate_deployments:
        # Validate certificate trust
        leaf_cert = cert_deployment.received_certificate_chain[0]
        if not cert_deployment.leaf_certificate_subject_matches_hostname:
            issues_with_certificates[
                "certificate_hostname_validation"
            ] = f"Certificate hostname validation failed for {leaf_cert.subject.rfc4514_string()}."
        if not cert_deployment.verified_certificate_chain:
            issues_with_certificates[
                "certificate_path_validation"
            ] = f"Certificate not path validation failed for {leaf_cert.subject.rfc4514_string()}."

        # Validate the public key
        public_key = leaf_cert.public_key()
        if isinstance(public_key, EllipticCurvePublicKey):
            deployed_key_algorithms.add("ecdsa")
            if public_key.curve.name not in mozilla_config.certificate_curves:
                # TODO(AD): Disable the check on the curves; not even Google and Cloudflare are compliant...
                pass
                # problems_with_certificates["certificate_curves"] = (
                #     f"Certificate curve is {public_key.curve.name},"
                #     f" should be one of {expected_mozilla_config.certificate_curves}."
                # )

        elif isinstance(public_key, RSAPublicKey):
            deployed_key_algorithms.add("rsa")
            if mozilla_config.rsa_key_size and public_key.key_size < mozilla_config.rsa_key_size:
                issues_with_certificates[
                    "rsa_key_size"
                ] = f"RSA key size is {public_key.key_size}, minimum allowed is {mozilla_config.rsa_key_size}."

        else:
            deployed_key_algorithms.add(public_key.__class__.__name__)

        deployed_signature_algorithms.add(leaf_cert.signature_algorithm_oid._name)

        # Validate the cert's lifespan
        leaf_cert_lifespan = leaf_cert.not_valid_after - leaf_cert.not_valid_before
        if leaf_cert_lifespan.days > mozilla_config.maximum_certificate_lifespan:
            issues_with_certificates["maximum_certificate_lifespan"] = (
                f"Certificate life span is {leaf_cert_lifespan.days} days,"
                f" should be less than {mozilla_config.maximum_certificate_lifespan}."
            )

    # TODO(AD): It's unclear whether the Mozilla profile/configs takes into accounts servers with multiple leaf certs
    #  What follows is my personal guess as to how it should work for multi-certs deployments...

    # Validate the public key algorithms
    # At least one of the Mozilla cert types should have been detected in the server's cert deployments
    found_cert_type = False
    for key_algorithm in mozilla_config.certificate_types:
        if key_algorithm in deployed_key_algorithms:
            found_cert_type = True
            break
    if not found_cert_type:
        issues_with_certificates["certificate_types"] = (
            f"Deployed certificate types are {deployed_key_algorithms},"
            f" should have at least one of {mozilla_config.certificate_types}."
        )

    # Validate the signature algorithms
    found_sig_algorithm = False
    for sig_algorithm in mozilla_config.certificate_signatures:
        if sig_algorithm in deployed_signature_algorithms:
            found_sig_algorithm = True
            break
    if not found_sig_algorithm:
        issues_with_certificates["certificate_signatures"] = (
            f"Deployed certificate signatures are {deployed_signature_algorithms},"
            f" should have at least one of {mozilla_config.certificate_signatures}."
        )

    # TODO(AD): Maybe add check for ocsp_staple but that one seems optional in https://ssl-config.mozilla.org/

    return issues_with_certificates



from enum import Enum
from typing import Dict, Type, TYPE_CHECKING, Set

from sslyze.plugins.elliptic_curves_plugin import SupportedEllipticCurvesImplementation


from sslyze.plugins.certificate_info.implementation import CertificateInfoImplementation
from sslyze.plugins.compression_plugin import CompressionImplementation
from sslyze.plugins.early_data_plugin import EarlyDataImplementation
from sslyze.plugins.fallback_scsv_plugin import FallbackScsvImplementation
from sslyze.plugins.heartbleed_plugin import HeartbleedImplementation
from sslyze.plugins.http_headers_plugin import HttpHeadersImplementation
from sslyze.plugins.openssl_ccs_injection_plugin import OpenSslCcsInjectionImplementation
from sslyze.plugins.openssl_cipher_suites.implementation import (
    Sslv20ScanImplementation,
    Sslv30ScanImplementation,
    Tlsv10ScanImplementation,
    Tlsv13ScanImplementation,
    Tlsv12ScanImplementation,
    Tlsv11ScanImplementation,
)
from sslyze.plugins.robot.implementation import RobotImplementation
from sslyze.plugins.session_renegotiation_plugin import SessionRenegotiationImplementation
from sslyze.plugins.session_resumption.implementation import SessionResumptionSupportImplementation

if TYPE_CHECKING:
    from sslyze.plugins.plugin_base import ScanCommandImplementation  # noqa: F401


class ScanCommand(str, Enum):
    CERTIFICATE_INFO = "certificate_info"
    SESSION_RESUMPTION = "session_resumption"
    SSL_2_0_CIPHER_SUITES = "ssl_2_0_cipher_suites"
    SSL_3_0_CIPHER_SUITES = "ssl_3_0_cipher_suites"
    TLS_1_0_CIPHER_SUITES = "tls_1_0_cipher_suites"
    TLS_1_1_CIPHER_SUITES = "tls_1_1_cipher_suites"
    TLS_1_2_CIPHER_SUITES = "tls_1_2_cipher_suites"
    TLS_1_3_CIPHER_SUITES = "tls_1_3_cipher_suites"
    TLS_COMPRESSION = "tls_compression"
    TLS_1_3_EARLY_DATA = "tls_1_3_early_data"
    OPENSSL_CCS_INJECTION = "openssl_ccs_injection"
    TLS_FALLBACK_SCSV = "tls_fallback_scsv"
    HEARTBLEED = "heartbleed"
    ROBOT = "robot"
    SESSION_RENEGOTIATION = "session_renegotiation"
    HTTP_HEADERS = "http_headers"
    ELLIPTIC_CURVES = "elliptic_curves"


class ScanCommandsRepository:
    @staticmethod
    def get_implementation_cls(scan_command: ScanCommand) -> Type["ScanCommandImplementation"]:
        return _IMPLEMENTATION_CLASSES[scan_command]

    @staticmethod
    def get_all_scan_commands() -> Set[ScanCommand]:
        return set(_IMPLEMENTATION_CLASSES.keys())


_IMPLEMENTATION_CLASSES: Dict[ScanCommand, Type["ScanCommandImplementation"]] = {
    ScanCommand.CERTIFICATE_INFO: CertificateInfoImplementation,
    ScanCommand.SESSION_RESUMPTION: SessionResumptionSupportImplementation,
    ScanCommand.SSL_2_0_CIPHER_SUITES: Sslv20ScanImplementation,
    ScanCommand.SSL_3_0_CIPHER_SUITES: Sslv30ScanImplementation,
    ScanCommand.TLS_1_0_CIPHER_SUITES: Tlsv10ScanImplementation,
    ScanCommand.TLS_1_1_CIPHER_SUITES: Tlsv11ScanImplementation,
    ScanCommand.TLS_1_2_CIPHER_SUITES: Tlsv12ScanImplementation,
    ScanCommand.TLS_1_3_CIPHER_SUITES: Tlsv13ScanImplementation,
    ScanCommand.TLS_COMPRESSION: CompressionImplementation,
    ScanCommand.TLS_1_3_EARLY_DATA: EarlyDataImplementation,
    ScanCommand.OPENSSL_CCS_INJECTION: OpenSslCcsInjectionImplementation,
    ScanCommand.TLS_FALLBACK_SCSV: FallbackScsvImplementation,
    ScanCommand.HEARTBLEED: HeartbleedImplementation,
    ScanCommand.ROBOT: RobotImplementation,
    ScanCommand.SESSION_RENEGOTIATION: SessionRenegotiationImplementation,
    ScanCommand.HTTP_HEADERS: HttpHeadersImplementation,
    ScanCommand.ELLIPTIC_CURVES: SupportedEllipticCurvesImplementation,
}
